import { Component, OnInit, DoCheck, OnDestroy, ViewChild, ViewChildren, QueryList } from "@angular/core";
import { ActivatedRoute } from '@angular/router';

import { Observer } from "jalhyd";

import { FormulaireService } from "../../services/formulaire/formulaire.service";
import { InternationalisationService } from "../../services/internationalisation/internationalisation.service";
import { FieldSet } from "../../formulaire/fieldset";
import { FormulaireDefinition } from "../../formulaire/definition/form-definition";
import { CalculatorResultsComponent } from "../../components/calculator-results/calculator-results.component";
import { Subscription } from "rxjs/Subscription";
import { FieldSetComponent } from "../field-set/field-set.component";
import { BaseComponent } from "../base/base.component";
import { CalculatorNameComponent } from "./calc-name.component";
import { FormulaireElement } from "../../formulaire/formulaire-element";
import { FieldsetContainer } from "../../formulaire/fieldset-container";
import { FieldsetContainerComponent } from "../fieldset-container/fieldset-container.component";

@Component({
    selector: 'hydrocalc',
    templateUrl: "./calculator.component.html",
    styles: [`
    .button_compute_ok {
        height: 3em;
        width: 30%;
    }

    .button_compute_err {
        height: 3em;
        width: 30%;
        color: red
    }

    #close-button {
        font-size: 1.5em;
    }
    `
    ]
})
export class GenericCalculatorComponent extends BaseComponent implements OnInit, DoCheck, OnDestroy, Observer {
    /**
     * liste des FieldSetComponent
     */
    @ViewChildren(FieldSetComponent)
    private _fieldsetComponents: QueryList<FieldSetComponent>;

    /**
     * liste des FieldsetContainerComponent
     */
    @ViewChildren(FieldsetContainerComponent)
    private _fieldsetContainerComponents: QueryList<FieldsetContainerComponent>;

    /**
     * composant d'affichage des résultats
     */
    @ViewChild(CalculatorResultsComponent)
    private resultsComponent: CalculatorResultsComponent;

    /**
     * composant "nom de la calculette"
     */
    @ViewChild(CalculatorNameComponent)
    private _calculatorNameComponent: CalculatorNameComponent;

    /**
     * formulaire affiché
     */
    private _formulaire: FormulaireDefinition;

    private _subscription: Subscription; // pour souscrire aux changements d'URL envoyés par le routeur

    /**
     * true si on a cliqué sur le bouton de lancement de calcul
     */
    private _computeClicked: boolean;

    /**
     * flag de validité gloable de la saisie
     * NB : la validité du bouton "calculer" dépend de la validité de la saisie dans l'UI et de celle indiquée par le formulaire.
     * La validité de l'UI comprend la forme (pas de chaîne alpha dans les champs numériques, etc..).
     * La validité formulaire comprend le domaine de définition des valeurs saisies.
     */
    private _isUIValid: boolean = false;

    /**
     * flag disabled du bouton "calculer"
     */
    private isCalculateDisabled: boolean = true;

    constructor(private intlService: InternationalisationService,
        private formulaireService: FormulaireService,
        private route: ActivatedRoute) {
        super();
    }

    private get formElements(): FormulaireElement[] {
        if (this._formulaire == undefined)
            return [];
        return this._formulaire.formElements;
    }

    /**
     * détermine si un FormulaireElement est du type FieldSet
     */
    private isFieldset(fe): boolean {
        return fe instanceof FieldSet;
    }

    /**
     * détermine si un FormulaireElement est du type FieldsetContainer
     */
    private isFieldsetContainer(fe): boolean {
        return fe instanceof FieldsetContainer;
    }

    private get hasForm() {
        return this._formulaire != undefined;
    }

    private get hasResults() {
        if (this.hasForm)
            return this._formulaire.hasResults;
        return false;
    }

    private get uitextTitre() {
        if (this.hasForm)
            return this.formulaireService.getLocalisedTitleFromCalculatorType(this._formulaire.calculatorType);

        return undefined;
    }

    private get uitextCalculer() {
        return this.intlService.localizeText("INFO_CALCULATOR_CALCULER");
    }

    private get uitextCloseDialogTitle() {
        return this.intlService.localizeText("INFO_CLOSE_DIALOGUE_TITRE");
    }

    private get uitextCloseDialogText() {
        return this.intlService.localizeText("INFO_CLOSE_DIALOGUE_TEXT");
    }

    private get uitextCloseDialogYes() {
        return this.intlService.localizeText("INFO_OPTION_YES");
    }

    private get uitextCloseDialogNo() {
        return this.intlService.localizeText("INFO_OPTION_NO");
    }

    ngOnInit() {
        this.intlService.addObserver(this);
        this.formulaireService.addObserver(this);
        this.subscribeRouter();
        this.formulaireService.updateLocalisation();
    }

    ngDoCheck() {
        if (this._formulaire != undefined)
            this.isCalculateDisabled = !(this._formulaire.isValid && this._isUIValid);
    }

    ngOnDestroy() {
        this.unsubscribeRouter();
        this.formulaireService.removeObserver(this);
        this.intlService.removeObserver(this);
    }

    private subscribeRouter() {
        // récupération des paramètres passés avec la route (URL)
        this._subscription = this.route.params.subscribe(params => {
            const uid: number = params['uid'];
            this.formulaireService.setCurrentForm(+uid);
        });
    }

    private unsubscribeRouter() {
        this._subscription.unsubscribe();
    }

    /*
     * gestion des événements clic sur les radios
     */
    private onRadioClick(info: any) {
        this._formulaire.onRadioClick(info);
    }

    private onCloseForm() {
        this.formulaireService.requestCloseForm(this._formulaire.uid);
    }

    /**
     * met à jour l'interface
     */
    // private updateUI() {
    //     // this.appRef.tick();
    //     // this.changeDetectorRef.detectChanges();  // provoque une détection des changements dans les contrôles
    //     // if (!this.changeDetectorRef['destroyed']) // pour éviter l'erreur "Attempt to use a destroyed view: detectChanges"
    //     //     this.changeDetectorRef.detectChanges();
    //     this.changeDetectorRef.markForCheck();
    // }

    /**
     * relit les valeurs dans l'interface et met à jour les NgParameter
     */
    private updateParametersFromUI() {
        this._fieldsetComponents.forEach(fsc => fsc.updateParametersFromUI());
        this._fieldsetContainerComponents.forEach(fscc => fscc.updateParametersFromUI());
    }

    private doCompute() {
        this.updateParametersFromUI();
        this._formulaire.doCompute();
        this._computeClicked = true;
    }

    private onCalcResultsViewChecked() {
        if (this._computeClicked) {
            this._computeClicked = false;
            this.scrollToResults();
        }
    }

    private scrollToResults() {
        const element = document.getElementById("resultsComp");
        if (element)
            element.scrollIntoView();
    }

    private getFieldsetStyleDisplay(id: string) {
        let isDisplayed: boolean = this._formulaire.isDisplayed(id);
        return isDisplayed ? "block" : "none";
    }

    private setForm(f: FormulaireDefinition) {
        if (this._formulaire != undefined)
            this._formulaire.removeObserver(this);
        this._formulaire = f;
        if (this._formulaire != undefined)
            this._formulaire.addObserver(this);
    }

    private updateFormulaireResults(uid: number) {
        if (this._formulaire.uid == uid)
            this.resultsComponent.updateView();
    }

    // interface Observer

    update(sender: any, data: any): void {
        if (sender instanceof InternationalisationService) {
            this.formulaireService.updateLocalisation();
        }
        else if (sender instanceof FormulaireService) {
            switch (data["action"]) {
                case "currentFormChanged":
                    const uid: number = +data["formId"];
                    this.setForm(this.formulaireService.getFormulaireFromId(uid));
                    this.resultsComponent.formulaire = this._formulaire;
                    this._calculatorNameComponent.model = this._formulaire;
                    break;
            }
        }
        else if (sender instanceof FormulaireDefinition) {
            switch (data["action"]) {
                case "resetForm": // réinitialisation du formulaire
                    var f = sender as FormulaireDefinition;
                    this.updateFormulaireResults(f.uid);
                    //     this.updateUI();
                    break;

                case "resultsUpdated":
                    f = sender as FormulaireDefinition;
                    this.updateFormulaireResults(f.uid);
            }
        }
    }

    /**
     * appelé après le 1er affichage du composant
     */
    protected afterFirstViewChecked() {
        this.updateUIValidity();
    }

    /**
     * calcul de la validité globale de la vue
     */
    private updateUIValidity() {
        this._isUIValid = false;

        if (this._fieldsetComponents != undefined)
            this._isUIValid = this._fieldsetComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    fieldset,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && fieldset.isValid;
                }
                // valeur initiale
                , this._fieldsetComponents.length > 0);

        if (this._fieldsetContainerComponents != undefined)
            this._isUIValid = this._isUIValid && this._fieldsetContainerComponents.reduce(
                // callback
                (
                    // accumulator (valeur précédente du résultat)
                    acc,
                    // currentValue (élément courant dans le tableau)
                    fieldsetContainer,
                    // currentIndex (indice courant dans le tableau)
                    currIndex,
                    // array (tableau parcouru)
                    array
                ) => {
                    return acc && fieldsetContainer.isValid;
                }
                // valeur initiale
                , true);
    }

    /**
     * réception d'un événement de validité d'un FieldSetComponent
     */
    private OnFieldsetValid() {
        this.updateUIValidity();
    }

    /**
     * réception d'un événement de validité d'un FieldsetContainerComponent
     */
    private onFieldsetContainerValid() {
        this.updateUIValidity();
    }

    /**
     * réception d'un événement de changement de valeur d'un input
     */
    private onInputChange() {
        this._formulaire.resetResults();
    }
}
