import { CalculatorType, ComputeNodeType, Structure, ParallelStructure, SessionNub } from "jalhyd";

import { FormulaireDefinition } from "../form-definition";
import { CalculatorResults } from "../../../results/calculator-results";
import { ServiceFactory } from "../../../services/service-factory";
import { ParamService } from "../../../services/param/param.service";
import { FormDefParamToCalculate } from "../form-def-paramcalc";
import { FormDefFixedVar } from "../form-def-fixedvar";
import { FormDefParallelStructures } from "../form-def-parallel-structures";
import { FormComputeParallelStructures } from "../form-compute-parallel-structures";
import { FormResultFixedVar } from "../form-result-fixedvar";
import { FieldsetContainer } from "../../fieldset-container";
import { FieldSet } from "../../fieldset";
import { SelectField } from "../../select-field";
import { NgParameter } from "../../ngparam";
import { FieldsetTemplate } from "../../fieldset-template";
import { FormulaireNode } from "../../formulaire-node";

export class FormulaireParallelStructure extends FormulaireDefinition {
    private _formFixedVar: FormDefFixedVar;

    private _formParallelStruct: FormDefParallelStructures;

    private _formParamCalc: FormDefParamToCalculate;

    private _formCompute: FormComputeParallelStructures;

    private _formResult: FormResultFixedVar;

    /**
     * id du select configurant le type d'ouvrage
     */
    private __ouvrageSelectId: string;

    constructor() {
        super(CalculatorType.ParallelStructure)
        this._formFixedVar = new FormDefFixedVar(this);
        this._formParamCalc = new FormDefParamToCalculate(this);
        this._formResult = new FormResultFixedVar(this, false);
        this._formParallelStruct = new FormDefParallelStructures();
        this._formCompute = new FormComputeParallelStructures(this, this._formParallelStruct, this._formResult);
    }

    private createStructNub(templ: FieldsetTemplate): SessionNub {
        const paramService: ParamService = ServiceFactory.instance.paramService;

        // valeurs par défaut de CalculatorType, ComputeNodeType, StructureType, LoiDebit
        // !!! attention !!! pour l'instant, il doit y avoir cohérence entre ces valeurs et celles du fichier de conf
        // cad valeur par défaut du 1er select (type d'ouvrage), du 2ème (loi de débit).
        // A terme, il faudrait analyser le fichier de conf (dépendances d'existence) pour déterminer automatiquement ces valeurs
        const params = FieldSet.makeDefaultProps(templ.calcTypeFromConfig, templ.defaultNodeTypeFromConfig);
        return paramService.createSessionNub(params);
    }

    /**
     * ajoute un Nub Structure
     * @param st structure à ajouter
     * @param after position après laquelle insérer la structure, à la fin sinon
     */
    private addStructureNub(st: Structure, after?: number) {
        this.parallelStructureNub.addStructure(st, after);
    }

    private get parallelStructureNub(): ParallelStructure {
        const params = {
            "calcType": CalculatorType.ParallelStructure,
            "nodeType": ComputeNodeType.None
        };
        return this.getSessionNub(params).nub as ParallelStructure;
    }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}): FieldSet {
        if (json["calcType"] === "Structure") {
            // indice après lequel insérer le nouveau FieldSet
            const after = data["after"];

            const res: FieldSet = new FieldSet(parent);
            const sn = this.createStructNub(data["template"]);
            this.addStructureNub(sn.nub as Structure, after);
            res.setSessionNub(sn, false);

            if (after !== undefined)
                parent.kids.splice(after + 1, 0, res);
            else
                parent.kids.push(res);

            this.resetResults();

            return res;
        }
        else
            return super.createFieldset(parent, json, data);
    }

    protected initParse() {
        this._formParamCalc.initParse();
    }

    protected parseOptions(json: {}) {
        // id du select configurant le type d'ouvrage
        this.__ouvrageSelectId = this.getOption(json, "ouvrageSelectId");
    }

    public afterParseFieldset(fs: FieldSet) {
        // si le FieldSet contient le select de type d'ouvrage
        if (this.__ouvrageSelectId) {
            const sel = fs.getFormulaireNodeById(this.__ouvrageSelectId);
            if (sel) {
                // on abonne le formulaire aux propriétés du FieldSet
                fs.properties.addObserver(this);
            }
        }
    }

    public moveFieldsetUp(fs: FieldSet) {
        if (fs.sessionNub.nub instanceof Structure) {
            // déplacement du nub
            this._paramService.moveStructureNubUp(fs.sessionNub);

            // déplacement du fieldset
            this.fieldsetContainer.moveFieldsetUp(fs);

            this.resetResults();
        }
        else super.moveFieldsetUp(fs);
    }

    public moveFieldsetDown(fs: FieldSet) {
        if (fs.sessionNub.nub instanceof Structure) {
            // déplacement du nub
            this._paramService.moveStructureNubDown(fs.sessionNub);

            // déplacement du fieldset
            this.fieldsetContainer.moveFieldsetDown(fs);

            this.resetResults();
        }
        else super.moveFieldsetDown(fs);
    }

    public removeFieldset(fs: FieldSet) {
        if (fs.sessionNub.nub instanceof Structure) {
            // suppression du nub
            this._paramService.deleteSessionNub(fs.sessionNub);

            // suppression du fieldset
            this.fieldsetContainer.removeFieldset(fs);

            this.resetResults();
        }
        else super.removeFieldset(fs);
    }

    protected completeParse(json: {}) {
        this._formParamCalc.parseOptions(json);
        this.subscribeFieldsetContainer();
    }

    /**
     * gestion du clic sur les radios "paramètre fixé, à varier, à calculer"    
     */
    public onRadioClick(info: string) {
        super.onRadioClick(info);
        this._formParamCalc.onRadioClick(info);
    }

    /**
     * @return une chaîne représentant le "contexte" courant (ici, combinaison type d'ouvrage-loi de débit)
     * @param fs FieldSet contenant les listes déroulantes type d'ouvrage et loi de débit
     */

    public resetResults() {
        this._formResult.resetResults();
    }

    public doCompute() {
        this._formCompute.doCompute();
    }

    public get hasResults(): boolean {
        return this._formResult.hasResults;
    }

    public get results(): CalculatorResults[] {
        return this._formResult.results;
    }

    private get fieldsetContainer(): FieldsetContainer {
        const n = this.getFormulaireNodeById("struct_container");
        if (n == undefined || !(n instanceof FieldsetContainer))
            throw new Error("l'élément 'struct_container' n'est pas du type FieldsetContainer");
        return n as FieldsetContainer;
    }

    /**
     * abonnement en tant qu'observateur du FieldsetContainer
     */
    private subscribeFieldsetContainer() {
        this.fieldsetContainer.addObserver(this);
    }

    /**
     * abonnement en tant qu'observateur des NgParameter des FieldSet contenus dans le FieldsetContainer
     */
    private subscribeStructureInputFields(fs: FieldSet) {
        for (const n of fs.allFormElements)
            if (n instanceof NgParameter)
                n.addObserver(this);
    }
    /**
     * abonnement en tant qu'observateur du SelectField des FieldSet contenus dans le FieldsetContainer
     */
    private subscribeStructureSelectFields(fs: FieldSet) {
        for (const n of fs.allFormElements)
            if (n instanceof SelectField)
                n.addObserver(this);
    }

    // interface Observer

    public update(sender: any, data: any) {
        if (sender instanceof FieldsetContainer)
            switch (data.action) {
                case "newFieldset":
                    this.reset();
                    this.subscribeStructureInputFields(data["fieldset"]);
                    this.subscribeStructureSelectFields(data["fieldset"]);
            }
        else if (sender instanceof FieldSet && data.action == "propertyChange") {
            switch (sender.id) {
                case "fs_ouvrage":
                    // this.replaceCurrentSessionNub(sender.properties);
                    const newNub = this.replaceSessionNub(sender.sessionNub, sender.properties);
                    sender.setSessionNub(newNub);
                    this.reset();
                    break;
            }
        }
    }
}
