import { CalculatorType, ComputeNodeType, ParamDefinition, LoiDebit, StructureType, Props, SessionNub, StructureProperties, Observer } from "jalhyd";

import { FormulaireElement } from "./formulaire-element";
import { Dependency } from "./dependency/dependency";
import { DependencyConditionType } from "./dependency/dependency-condition";
import { Field } from "./field";
import { CheckField } from "./check-field";
import { SelectField } from "./select-field";
import { NgParameter, ParamRadioConfig } from "./ngparam";
import { ServiceFactory } from "../services/service-factory";
import { ParamService } from "../services/param/param.service";
import { FormulaireDefinition } from "./definition/form-definition";
import { StringMap } from "../stringmap";
import { FieldsetContainer } from "./fieldset-container";
import { FormulaireNode } from "./formulaire-node";

export class FieldSet extends FormulaireElement implements Observer {
    /**
     * formulaire parent
     */
    private _parentForm: FormulaireDefinition;

    /**
     * SessionNub associé
     */
    private _sessionNub: SessionNub;

    /**
     * dictionnaire de traduction
     */
    private _localisation: StringMap;

    /**
     * fichier de configuration
     */
    private _jsonConfig: {};

    /**
     * propriétés déterminant l'état actuel du FieldSet (type de section, loi de débit, ...)
     */
    private _props: Props;

    constructor(parent: FormulaireNode) {
        super(parent);
        this._props = new Props();
    }

    public get sessionNub(): SessionNub {
        return this._sessionNub;
    }

    public setSessionNub(sn: SessionNub, update: boolean = true) {
        this._sessionNub = sn;
        if (update)
            this.updateFields();
    }

    private addField(f: Field) {
        if (f == undefined)
            throw new Error("FieldSet.addField() : argument incorrect (undefined)");
        this.kids.push(f);
    }

    public get hasInputs(): boolean {
        for (const f of this.kids)
            if (f instanceof NgParameter)
                return true;
        return false;
    }

    public getInput(i: number): NgParameter {
        let n = 0;
        for (let f of this.kids) {
            if (f instanceof NgParameter) {
                if (n == i)
                    return f;
                n++;
            }
        }
        return undefined;
    }

    public get isValid(): boolean {
        let res: boolean = true;
        for (const f of this.kids) {
            if (f instanceof Field)
                if (f.isDisplayed)
                    res = res && f.isValid;
        }
        return res;
    }

    private parse_check(json: {}): CheckField {
        let res: CheckField = new CheckField(this);
        res.parseConfig(json)
        return res;
    }

    private parse_select(json: {}): SelectField {
        let res: SelectField = new SelectField(this);
        res.parseConfig(json);
        res.addObserver(this);
        return res;
    }

    public get properties(): Props {
        return this._props;
    }

    public getPropValue(key: string): any {
        return this._props.getPropValue(key);
    }

    public setPropValue(key: string, val: any) {
        if (this._props.setPropValue(key, val, this)) {
            // si prop=type d'ouvrage, on prend une loi de débit compatible avec (spécifique aux ouvrages //) comme valeur par défaut
            if (key === "structureType") {
                if (!StructureProperties.isCompatibleValues(val, this.getPropValue("loiDebit")))
                    this.setPropValue("loiDebit", StructureProperties.findCompatibleLoiDebit(val));
            }
            // si prop=loi débit, on prend un type d'ouvrage compatible
            else if (key === "loiDebit")
                if (!StructureProperties.isCompatibleValues(this.getPropValue("structureType"), val))
                    this.setPropValue("structureType", StructureProperties.findCompatibleStructure(val));
        }
    }

    /**
     * valeurs par défaut pour StructureType, LoiDebit en fonction d'un ComputeNodeType
     */
    private static defaultProps(nodeType: ComputeNodeType): [StructureType, LoiDebit] {
        let structType: StructureType;
        let loiDebit: LoiDebit
        switch (nodeType) {
            case ComputeNodeType.StructureRectangle:
                structType = StructureType.VanneRectangulaire;
                loiDebit = LoiDebit.Cem88v;
                break;
        }
        return [structType, loiDebit];
    }

    /**
     * crée un objet Props servant de filtre pour sélectionner un Nub
     * @param calcType
     * @param nodeType 
     */
    public static makeDefaultProps(calcType: CalculatorType, nodeType: ComputeNodeType): Props {
        const res: Props = new Props();

        res.setPropValue("calcType", calcType);
        res.setPropValue("nodeType", nodeType);

        const p: [StructureType, LoiDebit] = FieldSet.defaultProps(nodeType);
        res.setPropValue("structureType", p[0]);
        res.setPropValue("loiDebit", p[1]);

        return res;
    }

    private getNubParamFromSymbol(symbol: string): ParamDefinition {
        return this._sessionNub.nub.getParameter(symbol);
    }

    /**
     * crée un input
     * @param json definition de l'input, extrait du fichier de conf de la calculette
     * @param node_type_filter filtre sur le type de noeud (input créé si undefined ou égal)
     * @param default_radio_config config du radio fixé/à varier/à calculer
     */
    private parse_input(json: {}, default_radio_config: string): NgParameter {
        const input_id: string = json["id"];
        const paramService: ParamService = ServiceFactory.instance.paramService;

        const calcType: CalculatorType = this.getPropValue("calcType");

        switch (input_id) {
            case "Pr":
                var res: NgParameter = paramService.createParameter(input_id, this);
                break;

            default:
                const nt: string = json["nodeType"];
                let nodeType: ComputeNodeType = nt == undefined ? this.getPropValue("nodeType") : ComputeNodeType[nt];

                if (nodeType === this.getPropValue("nodeType")) { // si le nodeType du paramètre est le même que celui du fieldset
                    try {
                        var nubParam: ParamDefinition = this.getNubParamFromSymbol(input_id);
                    }
                    catch (e) {
                    }
                    if (nubParam)
                        res = new NgParameter(nubParam, this);
                }
        }

        if (res)
            res.parseConfig(json, { "radioConfig": default_radio_config });

        return res;
    }

    private parseFields() {
        const fields = this._jsonConfig["fields"];
        for (const field_index in fields) {
            const field = fields[field_index];

            if (field["type"] === "input") {
                const default_radio_config = this._jsonConfig["option"];
                const param = this.parse_input(field, default_radio_config);
                if (param) // potentiellement undefined car certaines définitions de FieldSet comportent des paramètres qui ne sont pas tous affichés en même temps (cf. ouvrages //)
                    this.addField(param);
            } else if (field["type"] === "select") {
                const param = this.parse_select(field);
                this.addField(param);
            } else if (field["type"] === "check") {
                const param = this.parse_check(field);
                this.addField(param);
            }
        }
    }

    private clearFields() {
        for (const n of this.kids)
            n.removeObserver(this);

        this.clearKids();
    }

    public updateLocalisation(loc?: StringMap) {
        if (loc == undefined)
            loc = this._localisation;
        else
            this._localisation = loc;

        if (loc)
            super.updateLocalisation(loc);
    }

    /**
     * @param createOrUpdate true pour forcer la création d'un SessionNub
     */
    private updateFields() {
        this.clearFields();
        this.parseFields();
        this.parseDependencies(this._jsonConfig);
        this.updateLocalisation();

        // MAJ des selects avec les valeurs actuelles des propriétés
        // spécifique à chaque calculette, à revoir

        if (this._confId === "fs_ouvrage") {
            const sf1: SelectField = this.getFormulaireNodeById("select_ouvrage") as SelectField;
            const st: StructureType = this.getPropValue("structureType");
            const se1 = sf1.getSelectedEntryFromValue(st);
            sf1.setValue(se1);

            switch (st) {
                case StructureType.SeuilRectangulaire:
                    const sf2: SelectField = this.getFormulaireNodeById("select_loidebit1") as SelectField;
                    const se2 = sf2.getSelectedEntryFromValue(this.getPropValue("loiDebit"));
                    sf2.setValue(se2);
                    break;

                case StructureType.VanneRectangulaire:
                    const sf3: SelectField = this.getFormulaireNodeById("select_loidebit2") as SelectField;
                    const se3 = sf3.getSelectedEntryFromValue(this.getPropValue("loiDebit"));
                    sf3.setValue(se3);
                    break;
            }
        }
        else if (this._confId === "fs_section") {
            const sf: SelectField = this.getFormulaireNodeById("select_section") as SelectField;
            const nt: ComputeNodeType = this.getPropValue("nodeType");
            const se = sf.getSelectedEntryFromValue(nt);
            sf.setValue(se);
        }


        // fin MAJ selects

        this.applyDependencies(this._parentForm);
    }

    public parseConfig(json: {}, data?: {}) {
        this._jsonConfig = json;
        this._parentForm = data["parentForm"];

        this._confId = json["id"];

        const ct: string = json["calcType"];
        const calc_type: CalculatorType = ct == undefined ? this._parentForm.calculatorType : CalculatorType[ct];
        this.setPropValue("calcType", calc_type);

        const nt: string = json["nodeType"];
        const node_type: ComputeNodeType = nt == undefined ? ComputeNodeType.None : ComputeNodeType[nt];

        const dnt: string = json["defaultNodeType"];
        const default_node_type: ComputeNodeType = dnt == undefined ? ComputeNodeType.None : ComputeNodeType[dnt];

        if (nt !== undefined && dnt !== undefined)
            throw new Error("les champs 'nodeType' et 'defaultNodeType' ne doivent pas être définis en même temps")

        const ntype = dnt !== undefined ? default_node_type : node_type;

        this.setPropValue("nodeType", ntype);

        const st: string = json["defaultStructType"];
        if (st)
            this.setPropValue("structureType", StructureType[st])
        const ld: string = json["defaultLoiDebit"];
        if (ld)
            this.setPropValue("loiDebit", LoiDebit[ld])

        this.updateFields();
    }

    public parseDependencies(json: {}) {
        super.parseDependencies(json);

        for (const k1 in json) {
            if (k1 === "fields") {
                const fields = json[k1];
                for (const k2 in fields) {
                    const field = fields[k2];
                    switch (field["type"]) {
                        case "input":
                        case "select":
                        case "check":
                            for (const k of this.kids)
                                if (k.id == field["id"]) {
                                    k.parseDependencies(field);
                                    break;
                                }
                            break;
                    }
                }
            }
        }
    }

    public getNodeParameter(symbol: string): NgParameter {
        for (const p of this.kids)
            if (p instanceof NgParameter)
                if (p.isDisplayed && p.symbol === symbol)
                    return p;

        return undefined;
    }

    public getNodeParameterValue(symbol: string): number {
        const p = this.getNodeParameter(symbol);
        if (p == undefined)
            throw new Error(`FieldSet.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);

        switch (p.radioState) {
            case ParamRadioConfig.FIX:
                return p.getValue();

            case ParamRadioConfig.VAR:
            case ParamRadioConfig.CAL:
                return undefined;
        }
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField (qui doit être affiché)
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): string {
        for (const p of this.kids)
            if (p instanceof SelectField && p.isDisplayed && p.id === selectFieldId) {
                const value: string = p.getValue().value;
                return FormulaireElement.removePrefix(value, selectFieldId + "_");
            }
        return undefined;
    }

    // interface Observer

    update(sender: any, data: any) {
        if (data.action)
            switch (data.action) {
                case "select":
                    if (data.value.id.indexOf("select_section_") != -1)  // sections paramétrées
                        this.setPropValue("nodeType", data.value.value);
                    if (data.value.id.indexOf("select_ouvrage") != -1) // ouvrages parallèles
                        this.setPropValue("structureType", data.value.value);
                    else if (data.value.id.indexOf("select_loidebit1") != -1) // ouvrages parallèles
                        this.setPropValue("loiDebit", data.value.value);
                    else if (data.value.id.indexOf("select_loidebit2") != -1) // ouvrages parallèles
                        this.setPropValue("loiDebit", data.value.value);
                    else if (data.value.id.indexOf("select_resolution") != -1) // courbes de remous, méthode de résolution
                        this.setPropValue("methodeResolution", data.value.value);
                    else if (data.value.id.indexOf("select_target") != -1) // courbes de remous, variable à calculer
                        this.setPropValue("varCalc", data.value.value);
                    break;
            }
    }
}
