import { ParamDefinition, Pair, ParamDomain, ParamValues, ParamValueMode, ParamValueIterator } from "jalhyd";

import { InputField } from "./input-field";
import { Dependency } from "./dependency/dependency";
import { DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { FormulaireDefinition } from "./definition/form-definition";
import { ServiceFactory } from "../services/service-factory";
import { ApplicationSetupService } from "../services/app-setup/app-setup.service";
import { StringMap } from "../stringmap";
import { FormulaireNode } from "./formulaire-node";

export enum ParamRadioConfig {
    /**
     * pas de radio, paramètre modifiable à la main uniquement
     */
    FIX,

    /**
     * boutons radio "paramètre fixé" et "paramètre à varier"
     */
    VAR,

    /**
     * boutons radio "paramètre fixé", "paramètre à varier" et "paramètre à calculer"
     */
    CAL
};

/**
 * classe englobante de ParamDefinition (champs supplémentaires pour l'affichage, radio boutons, ...)
 */
export class NgParameter extends InputField {
    public unit: string;
    public radioConfig: ParamRadioConfig;

    /**
     * true si ce paramètre est celui par défaut dans un formulaire (cf. fichier de conf des calculettes, objet "options", champ "idCal")
     */
    public isDefault: boolean = false; // archi bug du langage ! si on relit cette propriété sans l'avoir modifiée entre-temps, elle vaut undefined !

    constructor(private _paramDef: ParamDefinition, parent: FormulaireNode) {
        super(parent);
    }

    get symbol(): string {
        return this._paramDef.symbol;
    }

    get domain(): ParamDomain {
        return this._paramDef.getDomain();
    }

    public getValue() {
        return this._paramDef.v;
    }

    public set confId(id: string) {
        if (this._confId != undefined)
            throw new Error(`NgParameter : l'identifiant de configuration est déjà défini (${this._confId})`);
        this._confId = id;
    }

    private get _paramValues() {
        return this._paramDef.paramValues;
    }

    public get paramDefinition() {
        return this._paramDef;
    }

    public get radioState() {
        switch (this._paramDef.paramValues.valueMode) {
            case ParamValueMode.SINGLE:
                return ParamRadioConfig.FIX;

            case ParamValueMode.LISTE:
            case ParamValueMode.MINMAX:
                return ParamRadioConfig.VAR;

            case ParamValueMode.CALCUL:
                return ParamRadioConfig.CAL;
        }
    }

    /**
     * notification envoyée après la modification de la valeur du paramètre
     */
    private notifyValueModified(sender: any) {
        this.notifyObservers(
            {
                "action": "ngparamAfterValue",
                "param": this,
                "value": this._paramDef.v
            }, sender
        );
    }

    /**
     * fixe la valeur du paramètre.
     * une notification préalable est envoyée pour laisser l'occasion aux objets liés de préciser le contexte
     * dans lequel cette valeur existe
     * @param sender 
     * @param val 
     */
    public setValue(sender: any, val: number) {
        this._paramDef.v = val;
        this.notifyValueModified(sender);
    }

    get isDefined(): boolean {
        return this._paramDef.isDefined;
    }

    public checkValue(val: number) {
        this._paramDef.checkValue(val);
    }

    public checkList(l: number[]) {
        for (let e of l)
            this.checkValue(e);
    }

    public get valueMode() {
        return this._paramDef.valueMode;
    }

    public set valueMode(m: ParamValueMode) {
        // undefined si on clique en dehors du select après l'avoir ouvert (cad sans avoir fait de sélection)
        // et au même niveau, cad à côté du bouton et non à côté du menu déroulant
        if (m != undefined)
            this._paramValues.valueMode = m;
    }

    /**
     * vérifie si un min/max est valide par rapport au domaine de définition
     */
    private isMinMaxDomainValid(v: number): boolean {
        if (v == undefined)
            return false;

        if (this._paramValues.valueMode == ParamValueMode.MINMAX)
            try {
                this.checkValue(v);
            }
            catch (e) {
                return false;
            }

        return true;
    }

    private checkMinMax(min: number, max: number): boolean {
        return this.isMinMaxDomainValid(min) && this.isMinMaxDomainValid(max) && (min < max);
    }

    public checkMin(min: number): boolean {
        return this.isMinMaxDomainValid(min) && (min < this._paramValues.max);
    }

    public checkMax(max: number): boolean {
        return this.isMinMaxDomainValid(max) && (this._paramValues.min < max);
    }

    public get isMinMaxValid(): boolean {
        return this.checkMinMax(this._paramValues.min, this._paramValues.max);
    }

    public get minValue() {
        return this._paramValues.min;
    }

    public set minValue(v: number) {
        this._paramValues.min = v;
    }

    public get maxValue() {
        return this._paramValues.max;
    }

    public set maxValue(v: number) {
        this._paramValues.max = v;
    }

    public checkStep(step: number): boolean {
        return this.isMinMaxValid && this._paramValues.stepRefValue.intervalHasValue(step);
    }

    public get stepRefValue(): Pair {
        return this._paramValues.stepRefValue;
    }

    public get stepValue() {
        return this._paramValues.step;
    }

    public set stepValue(v: number) {
        this._paramValues.step = v;
    }

    public get valueList() {
        return this._paramValues.valueList;
    }

    public set valueList(l: number[]) {
        this._paramValues.valueList = l;
    }

    private get isListValid(): boolean {
        if (this._paramValues.valueList == undefined)
            return false;

        for (let v of this._paramValues.valueList)
            try {
                this.checkValue(v);
            }
            catch (e) {
                return false;
            }
        return true;
    }

    private get isRangeValid(): boolean {
        switch (this._paramValues.valueMode) {
            case ParamValueMode.LISTE:
                return this.isListValid;

            case ParamValueMode.MINMAX:
                return this.checkStep(this._paramValues.step);
        }

        throw new Error(`"NgParameter.isRangeValid() : valeur ${ParamValueMode[this._paramValues.valueMode]} de ParamValueMode non prise en compte`);
    }

    private get isValueValid(): boolean {
        try {
            const v = this._paramDef.getValue();
            this._paramDef.checkValue(v);
            return true;
        }
        catch (e) {
            return false;
        }
    }

    public get isValid() {
        switch (this.radioState) {
            case ParamRadioConfig.FIX:
                return this.isValueValid;

            case ParamRadioConfig.VAR:
                return this.isRangeValid;

            case ParamRadioConfig.CAL:
                return true;

            case undefined:
                return false;
        }

        throw new Error(`"NgParameter.isValid() : valeur de ParamRadioConfig '${this.radioState}' (radioState) non prise en compte`);
    }

    private static getRadioConfig(s: string) {
        if (s == "fix")
            return ParamRadioConfig.FIX;

        if (s == "var")
            return ParamRadioConfig.VAR;

        if (s == "cal")
            return ParamRadioConfig.CAL;

        throw "invalid parameter radio configuration " + s;
    }

    public parseConfig(json: {}, data?: {}) {
        const appSetupService: ApplicationSetupService = ServiceFactory.instance.applicationSetupService;
        const radioConfig: string = data["radioConfig"];

        this._confId = json["id"];
        this.unit = json["unit"];
        if (this.symbol == "Pr")
            var val = appSetupService.computePrecision;
        else
            val = json["value"];
        if (val != undefined)
            this.setValue(this, +val);
        this.radioConfig = NgParameter.getRadioConfig(radioConfig);
        this.isDefault = false; // malgré le fait qu'il soit initialisé dans la déclaration de la classe NgParam à false, quand on relit sa valeur, il vaut undefined (merci Microsoft)
    }

    public verifiesDependency(d: Dependency): boolean {
        if (super.verifiesDependency(d))
            return true;

        switch (d.masterCondition.type) {
            case DependencyConditionType.HasValue:
                {
                    let mc: ValueDependencyCondition = <ValueDependencyCondition>d.masterCondition;
                    return this.getValue() === mc.value;
                }

            case DependencyConditionType.IsVariable:
                return this.radioState == ParamRadioConfig.VAR;

            default:
                throw "NgParameter.verifiesDependency() : type de condition '" + DependencyConditionType[d.masterCondition.type] + "' non pris en charge";
        }
    }

    public get valuesIterator(): ParamValueIterator {
        return this._paramValues.getValuesIterator();
    }

    public updateLocalisation(loc: StringMap) {
        super.updateLocalisation(loc);
        if (this.label == undefined) {
            const key: string = `INFO_GRANDEUR_${this.symbol.toUpperCase()}`;
            super.updateLocalisation(loc, key);
        }
    }
}
